#!/usr/bin/perl
# -w only while testing
# 
# mwiki (aka paxifixi)
#
# MAIL BASED CONTENT MANAGEMENT SYSTEM
#
# authors: Reyk Floeter <reyk@synack.de>,
#          Max-Gerd Retzlaff <m.retzlaff@gmx.net>,
#          Sebastian Horzela <jules@killall5.de>
# date:    $Date: 2002/02/07 00:51:31 $
# version: $Revision: 1.6 $
#
# This application is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version. 
#
# This application is distributed in the hope that it will be useful, 
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Library General Public License for more details.
#
# You should have received a copy of the GNU General Public
# License along with this library; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# 02111 USA.  

# changing to dir in which mwiki.pl is stored
BEGIN {
    (my $dir = $0) =~ s/^(.*)\/[^\/]*?$/$1/;
    chdir( $dir );
}

###################################################################
# default settings, do() config file, modules, usage information
#
#default# $config_dir = "/etc/mwiki";
$config_dir = "../etc/mwiki";
$config_file = $config_dir."/mwiki.conf";
# The config file is a piece of perl code now.
# You have (at least) to define:
#    - $top_dir
#    - $www_dir
#    - %types
#    - $mwiki_addr
# You should also define:
#    - $main_title
#    - $sub_title
#    - $index_title
#    - $default_body_file

# modules
use structuredText;
use HelpfulFunctions; 
use MailParser;
use generateHTML;
use GnuPGverify;
$GnuPG_valid_IDs_file = $config_dir."/valid_IDs.conf";
GnuPGverify->load_valid_IDs( $GnuPG_valid_IDs_file );
use Mail::Mailer;

$use_pointer = "yes";
$dir_perm_public   = 0755;
$file_perm_public  = 0644;
$file_perm_private = 0600;

### default values that should be overwritten in your config file (see above)
# now testing whether $top_dir is defined, so the next two lines are uncommented
# $top_dir = "nil";  
# $www_dir = "nil";  
%types = ();
$mwiki_addr = "mwiki\@localhost";

$main_title = "Contents";
$sub_title = "Contents of";
$index_title = "Index";

$default_body = "\n<h2>Welcome to mwiki</h2>\n<em>MAIL BASED CONTENT MANAGEMENT SYSTEM</em>\n";
$default_body_file = ''; # -> not present
###

# execute config file 
# this has to happen AFTER the assignment of the default values
do $config_file;

if ( (!defined $top_dir) or (!defined $www_dir) ) {
    mdie(1, "ERROR: Invalid configuration! Check \"$config_file\".\n");
}

generateHTML->set_www_dir($www_dir); # has to be invoked after "do $config_file"
file_as_string( \$default_body, $default_body_file );

$latest_news_file = "$top_dir/latest_news.inc";
$index_inc_file   = "$top_dir/index.inc";
$pointer_inc_file = "$top_dir/pointer.inc";

# prepare usage information
$usage = <<"END_OF_USAGE";

usage:\tmwiki.pl -[m|M|i|p|r|R|S]

\t-m  process message from stdin
\t-M  process message from stdin and generate pointers and indexes
\t-i  generate indexes
\t-p  make pointers
\t-r  re-generate html content
\t-R  generate indexes and pointers and re-generate html content
\t-S  create subdirectories and (empty) include files in \$top_dir,
\t      then generate indexes and pointers and re-generate html content
\t      (IOW: create a skeleton page / this is non-destructive)

\t note: You would normally use only the big-letter options.

status:
\t\$top_dir:\t$top_dir
\t\$www_dir:\t$www_dir
\t\$config_file:\t$config_file

END_OF_USAGE



###################################################################
# parse program parameters
if ($#ARGV != 0) {
    mdie(1, $usage);
}

if ($ARGV[0] eq "-m" or $ARGV[0] eq "-M") {
    my $stdin = '';
    {
	local $/ = undef;
	$stdin = <STDIN>;
    }
    parse_mail($stdin, \%types);
    if ($ARGV[0] eq "-M") {
	make_pointers();
	generate_indexes(\%types);
    }
} elsif ($ARGV[0] eq "-i") {
    generate_indexes(\%types);
} elsif ($ARGV[0] eq "-r") {
    generate_content(\%types);
} elsif ($ARGV[0] eq "-p") {
    make_pointers();
} elsif ($ARGV[0] eq "-R") {
    generate_indexes(\%types);
    make_pointers();
    generate_content(\%types);
} elsif ($ARGV[0] eq "-S") {
    create_skeleton(\%types);
    generate_indexes(\%types);
    make_pointers();
    generate_content(\%types);
} else {
    mdie(1, $usage);
}



###################################################################
# generate subdirectories and (empty) include files in $top_dir
# create_sekelton() is non-destructive. This meas that you can
# add categories (see %types) to an existing mwiki installation.
sub create_skeleton {
    my %types = %{$_[0]};

    my %created = ();
    foreach $dir (values %types) { # you could also use unique_array()
	mkdir( "$top_dir/$dir" ) unless $created{$dir}++;
	chmod( $dir_perm_public, "$top_dir/$dir" );
    }

    string_to_file(">".$latest_news_file, '', $file_perm_private);
    string_to_file(">".$index_inc_file, '', $file_perm_private);
    string_to_file(">".$index_inc_file, '', $file_perm_private);
    string_to_file(">".$pointer_inc_file, '', $file_perm_private);

    return 0;
}



###################################################################
# generates directory indexes
sub generate_indexes {
    my %types = %{$_[0]};

    my $latest_news = '';
    file_as_string( \$latest_news, $latest_news_file );

    # get subdirectories, preserve sorting of keys of %types
    my @dirs = unique_array(values %types);
    
    # Okay, now let's sort it. values(%hash) has got no "sorting" in any case...
    @dirs = sort @dirs;

    my $main .= "\t<a href='".$www_dir."/index.html'>".$main_title."</a><p>\n";
    foreach $cwd (@dirs) {
	$main .= "\t<li><a href=\"$www_dir/$cwd/index.html\">$cwd</a></li>\n";
    }
 
    if ($use_pointer eq "yes") {
	$main .= "\t<p><li><a href=\"".$www_dir."/pointer.html\">pointer</a></li>";
    }

    # write MAIN MENU
    string_to_file("$index_inc_file", $main, $file_perm_private);

    # write $top_dir."/index.html"
    string_to_file($top_dir."/index.html",
		   generateHTML->whole_page($main_title, '', $index_title, $main,
					    $main_title, $default_body,
					    $latest_news),
		   $file_perm_public);

    # generate indexes for subdirectories
    foreach $cwd (@dirs) {
	opendir(MYDIR, $top_dir."/".$cwd);
    
	# write index for subdirectory
	my $htmltext = '';

	while($filename = readdir(MYDIR)) {
	    if ($filename =~ /\.txt$/) {

		my $rawmail = '';
		file_as_string( \$rawmail, $top_dir."/".$cwd."/".$filename );
		
		my $mail = MailParser->autodetectingnew($rawmail);

		my $from = $mail->get_sender();
		my (undef, undef, undef, $title) = $mail->parse_subject();
		my $date = $mail->get_header_field( 'Date' );

		(my $filename_html = $filename ) =~ s/(.*)\.txt/$1\.html/;

		$htmltext .= "\t<li><a href='".$filename_html."'>".$title."</a>&nbsp\n";
		$htmltext .= "<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;\n";
		$htmltext .= "(by <font size='-1'><b>".$from."</b> on ".$date.")&nbsp; ";
		$htmltext .= "<a href='".$filename."'>[src]</a></font></li>\n";
	    }
	}

	string_to_file($top_dir."/".$cwd."/index.html", 
		       generateHTML->whole_page($cwd, '', $index_title, $main,
						"$sub_title $cwd",
						"<ul>$htmltext</ul>", $latest_news),
		       $file_perm_public);
	closedir(MYDIR);
    }
    return 0;
}



###################################################################
# create $topdir."/pointer.html"
sub make_pointers
{
    my $index = '';
    file_as_string( \$index, $index_inc_file );

    my $pointers = '';
    file_as_string( \$pointers, $pointer_inc_file);

    $pointers = join("\n", unique_array( split("\n", $pointers) ) );

    my $htmltext = structuredText2HTML("\n".$pointers);

    string_to_file($top_dir."/pointer.html",
		   generateHTML->whole_page("Pointers", '', $index_title, $index,
					    "Pointers", $htmltext, 
					    scalar localtime() ),
		   $file_perm_public );

    return 0;
}



###################################################################
# re-creates content files
sub generate_content
{
    my %types = %{$_[0]};
    my $filename;

    ## delete pointer file
    # unlink $pointer_inc_file;
    #  this would not work as intended
    #  because links of refused mails cannot be regained
    #  -> new, third option for parse_mail()
    
    my @dirs = unique_array(values %types);

    foreach $cwd (@dirs) {
	opendir(MYDIR, $top_dir."/".$cwd);
 
	while($filename = readdir(MYDIR)) {
	    if ($filename =~ /\.txt/) {
		my $input = '';
		file_as_string( \$input, $top_dir."/".$cwd."/".$filename );
		parse_mail($input, \%types, "re-generation (do not search for links)");
	    }
	}
	
	closedir(MYDIR);
    }
    return 0;
}



###################################################################
# processes incomming messages
sub parse_mail
{
    my $rawmail = $_[0];
    my %types = %{$_[1]};
    my $regeneration = defined($_[2]) ? '0' : '1';

    if ($regeneration) {
	my @pointer = $rawmail =~ m/(\"[^\"]*?\"\s\((?:http|https|ftp|news|telnet):.*?\))/gis;
	
	foreach $link (@pointer) { # correct messed links, strip =\n
	    $link =~ s/\=\n//g;
	}

        # alternative:
	# my @pointer = ();
	# while ( $rawmail =~ m/(\"[^\"]*?\"\s\((http|https|ftp|news|telnet|mailto):.*?\))/gis ) {
	# push( @pointer, $1);
	# # print STDERR "->$1<-\n";
        # }
	# print join("\n", @pointer)."\n";

	string_to_file(">".$pointer_inc_file, 
		       join('', map("- $_\n", @pointer) ), # join() is needed!
		       $file_perm_private);

	make_pointers(); # <- is a workaround, cf. TODO file
    }

    my $mail = MailParser->autodetectingnew($rawmail);

    my $date = $mail->get_header_field( 'Date' );
    my $from = $mail->get_sender();

    my ($command, $type, $mid, $title) = $mail->parse_subject();
    # print "-$command-$type-$mid-$title-\n";

    # send digest
    if ( $command eq 'SENDDIGEST' ) {
	send_digest( $type, $mid, $title, \%types );
	return 0; # go out of parse_mail()
    }

    # process digest
    if ( $command eq 'DIGEST' ) {
	process_digest( $rawmail, \%types );
	return 0; # go out of parse_mail()
    }

    $mid = $mail->get_mid() if ( $mid eq '' );
    # use a random filename for missing Message-IDs...
    $mid = time()."-".rand(1).'@no.where' if ( $mid eq '');

    $mail->replace_subject_and_mid( $command, $type, $mid, $title );

    mdie(0,"ERROR: unsupported mail type (Don't support messages of type $type.)\n") if (!defined($types{$type})); # ($types{$type} eq "");    

    # convert structured Text to HTML
    $htmlmail = structuredText2HTML( $mail->get_content() );
    mdie(1,"ERROR: Couldn't transform structured Text to html.\n") if ($htmlmail eq ""); # should not happen... ;-)

    $filename = $top_dir."/".$types{$type}."/mwiki-".$mid;

    my $index = "";
    file_as_string( \$index, $index_inc_file);

    if ($regeneration) {
	# write last updated information - new: only when a new mail is processed
	string_to_file($latest_news_file,
		   "<em><a href=\"$www_dir/".$types{$type}."/index.html\">".$types{$type}."</a>-&gt;<a href=\"$www_dir/".$types{$type}."/mwiki-$mid.html\">".$title."</a></em>, ".localtime(),
		       $file_perm_private );
    }

    my $special = "<p/ ><font size='-1'>";
    $special.= "[<a href='mwiki-".$mid.".txt'>view mail source</a>]&nbsp;\n";
    $special.= "[<a href='mailto:".$mwiki_addr."?subject=REPLACE: ".$type.": [".$mid."] ".$title."'>replace page</a>]&nbsp;\n";
    $special.= "[<a href='mailto:".$mwiki_addr."?subject=DELETE: ".$type.": [".$mid."] ".$title."'>delete page</a>]&nbsp;\n";
    $special.= "$date</font>";


    if ($command eq "DELETE") {
	unlink "$filename.html", "$filename.txt";
    } else {
	# write HTML'ized output
	string_to_file($filename.".html",
		       generateHTML->whole_page($title, '', $index_title, $index,
			    "<a href=\"$www_dir/".$types{$type}."/index.html\">".
			    $types{$type}." -&gt;</a> ".$title.
			    " &nbsp;<i><small>(by $from)</small></i>",
			    $htmlmail. $special, scalar localtime() ),
		       $file_perm_public );
	# write mail source
	# string_to_file($filename.".txt", $rawmail, $file_perm_public );
	string_to_file($filename.".txt", $$mail, $file_perm_public );
    }

    return 0;
}



###################################################################
# generates directory indexes
sub send_digest {
    my ($type, $mid, $title) = @_[0..2];
    my %types = %{$_[3]};

    my $digest = '';

    # get subdirectories, preserve sorting of keys of %types
    my @dirs = unique_array(values %types);

    # append mails to digest
    foreach $cwd (@dirs) {
	if ( $type eq 'MWIKIALL' or $cwd eq $types{$type} ) {
	    opendir(MYDIR, $top_dir."/".$cwd);
    
	    while($filename = readdir(MYDIR)) {
		if ($filename =~ /\.txt$/) {
		    my $rawmail = '';
		    file_as_string( \$rawmail, $top_dir."/".$cwd."/".$filename );
		    
		    $digest .= $rawmail."\n\n";
		}
	    }
	    closedir(MYDIR);
	}
    }

    # mask ^From\ 's
    $digest =~ s/(^|\n)(\>*From\ )/$1>$2/g;

    # send digest as mail to $mid
    $mid =~ s/-at-/\@/;
    eval {
	my $mailer = Mail::Mailer->new();
	$mailer->open( { "From"    => $mwiki_addr,
			 "To"      => $mid,
			 "Subject" => "DIGEST: $type: your title was: $title"
#			 "Content-Type" => "text/plain;"
			 } ) or mdie(1, "ERROR: Couldn't open \$mailer: $!.\n");
	print $mailer $digest;
	$mailer->close() or mdie(1, "ERROR: Couldn't close \$mailer: $!.\n");
    };
    if ($@) {
	mdie(1, "ERROR: Couldn't send digest to $mid. Eval error: $@.\n");
    }

    return 0;
}



###################################################################
# processes incomming messages
sub process_digest {
    my $rawmail = $_[0];
    my %types = %{$_[1]};

    my (undef, $content) =  split/\n\n/, $rawmail, 2;

    # unmask ^From\ 's
    $content =~ s/(^|\n)(\>)(\>*From\ )/$1$3/g;

    $content =~ s/(\n\nFrom\ )/$1From\ /g;
    my @mails = split("\n\n+From\ ", $content);

    foreach $mail (@mails) {
	parse_mail($mail, \%types);
    }	    

    return 0;
}



###################################################################
exit 0;
